#include <stdlib.h>
#include <stdio.h>

#define MAGIC_NUMBER 0xABCDEF42
#define HEAP_LIMIT (64 * 1024 * 1024)

typedef struct
{
    unsigned int magicNumber;
    unsigned int imageSize;
    unsigned int nextHash;
    unsigned int specialOops;
} ImageHeader;

void
error (const char * message)
{
    fprintf (stderr, "Slate: %s\n", message);
    exit (EXIT_FAILURE);
}

#include "vm.h"

void
growMemoryBy (struct ObjectHeap * heap, unsigned long int delta)
{
    if (heap -> memoryEnd + delta > heap -> memoryLimit)
      error ("Attempted to grow heap beyond limits");
    
    heap -> memoryEnd += delta;

    fprintf (stderr, "Slate: Growing heap to %d bytes.\n", (unsigned) heap -> memoryEnd - (unsigned) heap -> memory);
}

void
shrinkMemoryBy (struct ObjectHeap * heap, unsigned long int delta)
{
    if (heap -> memoryEnd - delta <= (unsigned long int) heap -> memory)
      error ("Attempted to shrink heap to zero size");

    heap -> memoryEnd -= delta;

    fprintf (stderr, "Slate: Shrinking heap to %d bytes.\n", (unsigned) heap -> memoryEnd - (unsigned) heap -> memory);
}

int
main (int argc, char ** argv)
{
    FILE * file;
    ImageHeader header;
    struct ObjectHeap heap;
    
    if (argc < 2)
    {
        fprintf (stderr, "Usage: %s <image>\n", argv [0]);
        return EXIT_FAILURE;
    }

    file = fopen (argv [1], "rb");
    if (file == NULL)
      error ("Could not open image file");

    if (fread (& header, sizeof (ImageHeader), 1, file) != 1 ||
        header.magicNumber != MAGIC_NUMBER ||
        header.imageSize >= HEAP_LIMIT)
    {
        fclose (file);
        error ("Invalid image header");
    }

    heap.memory = (unsigned long int *) malloc (HEAP_LIMIT);
    if (heap.memory == NULL ||
        fread (heap.memory, 1, header.imageSize, file) != header.imageSize)
    {
        fclose (file);
        error ("Could not read image heap");
    }

    fclose (file);

    heap.memoryEnd = (unsigned long int) heap.memory + header.imageSize;
    heap.memoryLimit = (unsigned long int) heap.memory + HEAP_LIMIT;
    heap.lastHash = header.nextHash;
    heap.specialObjectsOop = header.specialOops;

    CurrentMemory = & heap;

    PSObjectHeap_initializeWithShift_ (CurrentMemory, (unsigned long int) heap.memory);
    
    PSInterpreter_interpret ((struct Interpreter *) PSObjectHeap_specialAt_ (CurrentMemory, InterpreterObject));

    return EXIT_SUCCESS;
}

